/**********************************************************************
 * client.c                                                           *
 *                                                                    *
 * Client Main Code                                                   *
 *                                                                    *
 * Brian Hoffman                                                      *
 * bhoffman@alum.mit.edu                                              *
 * ICFP Programming Contest 2002                                      *
 *                                                                    *
 **********************************************************************/

/******************************
 * Include Files              *
 ******************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <netdb.h>

#include "client_defs.h"
#include "client_data.h"

/******************************
 * Global Variables           *
 ******************************/
static Game_state game_state;
static Map_node   game_map[MAX_MAP_SIZE_X][MAX_MAP_SIZE_Y];

/******************************
 * Local Defines              *
 ******************************/

/******************************
 * Local Function Prototypes  *
 ******************************/
/* Init Functions
 */
static int  load_map(void);
static int    load_map_line(char *line, int width, int line_num);
static int    process_map(void);
static int  load_cfg(void);

/* Game Functions
 */
static int  game_init(void);
static int  game_play(void);

static int    get_package_list(void);
static int      package_add_entry(Package_entry **head,
                                  int id, int  x, int  y, int  wgt);
static int      package_remove_entry(Package_entry **head,
                                  int id, int *x, int *y, int *wgt);
static int      package_get_info( Package_entry  *head,
                                  int id, int *x, int *y, int *wgt);
static void     package_delete(Package_entry **head);

static int    issue_command(void);
static int      find_closest_home(int *x, int *y);
static int      plan_path(int s_x, int s_y, int d_x, int d_y, Path_entry **p);
static int        goal_dist_est(int s_x, int s_y, int d_x, int d_y);
static int      follow_path(int r_x, int r_y, Path_entry **path, char *cmd);

static int        path_length(     Path_entry     *head);
static int        path_push_node(  Path_entry    **head, Map_node  *node);
static int        path_pop_node(   Path_entry    **head, Map_node **node);
static void       path_delete(     Path_entry    **head);

static int        pque_push_node(  Pf_pque_entry **head, Map_node  *node);
static int        pque_pop_node(   Pf_pque_entry **head, Map_node **node);
static int        pque_find_node(  Pf_pque_entry  *head, Map_node  *node);
static int        pque_is_empty(   Pf_pque_entry  *head);
static void       pque_delete(     Pf_pque_entry **head);

static int        list_push_node(  Pf_list_entry **head, Map_node  *node);
static int        list_remove_node(Pf_list_entry **head, Map_node  *node);
static int        list_find_node(  Pf_list_entry  *head, Map_node  *node);
static void       list_delete(     Pf_list_entry **head);

static int    process_server_resp(void);
static void cleanup_exit(int status);

/* Socket Functions
 */
static int  client_connect(char *host_str, char *port_str);
static int  socket_write(char *data, int num_bytes);
static int  socket_read(char *data, int num_bytes);
static int  socket_fgets(char *data, int max_bytes);
static int  socket_fgets_unlim(char **pbuffer);

/* Debug Functions
 */
static void print_map(void);
static void print_map_links(void);
static void print_map_packages(void);
#ifdef DEBUG_PF
static void print_path(Path_entry *head);
#endif
#ifdef DEBUG_PKG2
static void print_package(Package_entry *head);
#endif

/***************************************************
 * main()                                          *
 *                                                 *
 ***************************************************/
int main(int argc, char *argv[])
{
  int status;

  /* Check input arguments
   *
   * Expected arguments are:
   * 1) Program Name (Unix Default)
   * 2) Hostname or IP address of the game server
   * 3) Port Number of the game server
   */
  if (argc != 3)
    {
      printf("usage: %s : <game server> <port>\n", argv[0]);
      exit(OK);
    }

  /* Connect to the game server
   */
  status = client_connect(argv[1], argv[2]);
  if (status != OK)
    {
      fprintf(stderr, "main(): Error in client_connect().\n");
      cleanup_exit(ERROR);
    }

  printf("main(): Connected to server.\n");

  /* Game Init
   */
  status = game_init();
  if (status != OK)
    {
      fprintf(stderr, "main(): Error in game_init().\n");
      cleanup_exit(ERROR);
    }

  printf("main(): game_init() complete.\n\n");

  /* Game Play
   */
  status = game_play();
  if (status != OK)
    {
      fprintf(stderr, "main(): Error in game_play().\n");
      cleanup_exit(ERROR);
    }

  printf("main(): Game Complete.\n");

  print_map_packages();

  cleanup_exit(OK);

  /* Avoid compiler warning
   */
  exit(OK);
}


/***************************************************
 * cleanup_exit()                                  *
 *                                                 *
 ***************************************************/
static void cleanup_exit(int status)
{
  /* Clean up allocated memory
   */
  package_delete( &(game_state.pkg_unpicked) );
  package_delete( &(game_state.pkg_picked)   );
  path_delete(    &(game_state.active_path)  );

  /* Close Open Socket Connections
   */
  if (game_state.socket_open)
    {
      close(game_state.socket);
    }
  

  exit(status);
}


/***************************************************
 * game_init()                                     *
 *                                                 *
 ***************************************************/
static int game_init(void)
{
  int  status;
  char start_str[] = "Player\n";

  /* Handle the initialization phase of the Game
   *
   * The initialization phase consists of 4 steps:
   *
   * 1) Client: send the string "Player"
   * 2) Server: send board [the Game board]
   * 3) Server: send player's configuration
   * 4) Server: send initial server response
   *            (with the initial location of all players)
   */
  status = socket_write(start_str, strlen(start_str));
  if (status != OK)
    {
      fprintf(stderr, "game_init(): Error in socket_write().n\n");
      return(ERROR);
    }

  printf("game_init(): Sent Player.\n");

  status = load_map();
  if (status != OK)
    {
      fprintf(stderr, "game_init(): Error in load_map().\n");
      return(ERROR);
    }

  printf("game_init(): Loaded map.\n");

  status = load_cfg();
  if (status != OK)
    {
      fprintf(stderr, "game_init(): Error in load_cfg().\n");
      return(ERROR);
    }

  printf("game_init(): Loaded cfg.\n");

  status = process_server_resp();
  if (status != OK)
    {
      fprintf(stderr, "game_init(): Error in process_server_resp().\n");
      return(ERROR);
    }

  printf("game_init(): Got initial response.\n");

  /* Update Game State
   */
  game_state.alive        = TRUE;
  game_state.weight       = 0; /* Currently carrying nothing */
  game_state.score_guess  = 0;
  game_state.pkg_picked   = NULL;
  game_state.pkg_unpicked = NULL;

  return(status);
}


/***************************************************
 * load_map()                                      *
 *                                                 *
 ***************************************************/
static int load_map(void)
{
  char   tmp[11];     /* "1000 1000\n\0" */
  char  *ptmp;
  char **stringp;
  char  *token;
  int    map_size_x;
  int    map_size_y;
  char  *line_buffer;
  int    num_bytes;
  int    status;
  int    i;

  /* Clear all of map memory
   */
  num_bytes = MAX_MAP_SIZE_X * MAX_MAP_SIZE_Y * sizeof(Map_node);
  memset((char *) &(game_map[0][0]), 0x0, num_bytes);

  /* The board is transmitted in the following format:
   *
   * 1) Server: board dimensions
   * 2) Server: board row 1 ...
   * 3) Server: ... other rows ...
   * 4) Server: board row n
   *
   * The dimensions are two integers, the first one indicating the
   * width (i.e. the number of columns), and the second one the height
   * (i.e. the number of rows) on the board.
   *
   * Once the dimensions are transmitted, the server sends one line
   * for every row of the board. The first line corresponds to the row
   * with vertical coordinate 1, i.e. the board is sent from south to
   * north.
   *
   * Every row consists of a number of characters, indicating the type
   * of tile at the corresponding position in the row. 
   *
   * The first character corresponds to horizontal position 1,
   * i.e. each row is sent from west to east. This is the encoding for
   * the different tile types:
   *
   * plain             . (a dot)  
   * lethal (water)    ~ (a tilde)  
   * impassable (wall) # (a hash)  
   * home base         @ (an at)  
   */
  status = socket_fgets(tmp, 11);
  if (status != OK)
    {
      fprintf(stderr, "load_map(): Error in socket_fgets(), ");
      return(ERROR);
    }

  ptmp       = &(tmp[0]);
  stringp    = &(ptmp);

  token      = strsep(stringp, " ");
  if (token == NULL)
    {
      fprintf(stderr, "load_map(): Unexpected NULL token.\n");
      fprintf(stderr, "line: __LINE__\n");
      return(ERROR);
    }
  map_size_x = atoi(token);

  token      = strsep(stringp, " ");
  if (token == NULL)
    {
      fprintf(stderr, "load_map(): Unexpected NULL token.\n");
      fprintf(stderr, "line: __LINE__\n");
      return(ERROR);
    }
  map_size_y = atoi(token);

  printf("load_map(): Got map size: %d x %d.\n", map_size_x, map_size_y);

  
  /* Allocate memory to read each line of the map
   * (allocate 1 extra byte for the \n character)
   */
  line_buffer = (char *) malloc((map_size_x+1) * sizeof(char));
  if (line_buffer == NULL)
    {
      perror("malloc(): Error ");
      fprintf(stderr, "load_map(): Error in malloc().\n");
      return(ERROR);
    }

  /* Read the map
   */
  status = OK;
  for (i = 1; i <= map_size_y; i++)
    {
      /* Make sure to read the \n character
       */
      status = socket_read(line_buffer, (map_size_x+1) * sizeof(char));
      if (status != OK)
        {
          fprintf(stderr, "load_map(): Error in socket_read().\n");
          free(line_buffer);
          return(ERROR);
        }

      /* Replace the \n with \0 for printing
       */
      line_buffer[map_size_x] = '\0';
      printf("%s\n", line_buffer);

      /* Load the data into the map nodes
       */
      status = load_map_line(line_buffer, map_size_x, i);
      if (status != OK)
        {
          fprintf(stderr, "load_map(): Error in load_map_line().\n");
          free(line_buffer);
          return(ERROR);
        }
    }

  /* Free allocated memory
   */
  free(line_buffer);


  /* Fill in the edges (Update the map size)
   */
  map_size_x += 2;
  map_size_y += 2;

  for (i = 0; i < map_size_x; i++)
    {
      game_map[0][i].type            = TERRAIN_WALL;
      game_map[map_size_y-1][i].type = TERRAIN_WALL;
    }
  for (i = 0; i < map_size_y; i++)
    {
      game_map[i][0].type            = TERRAIN_WALL;
      game_map[i][map_size_x-1].type = TERRAIN_WALL;
    }

  /* Save map size into game state
   */
  game_state.map_size_x = map_size_x;
  game_state.map_size_y = map_size_y;

  /* Print the map
   */
  print_map();


  /* Process the Map Nodes
   */
  status = process_map();
  if (status != OK)
    {
      fprintf(stderr, "load_map(): Error in process_map().\n");
      return(ERROR);
    }

  /* Print the map links
   */
  print_map_links();

  return(OK);
}


/***************************************************
 * load_map_line()                                 *
 *                                                 *
 ***************************************************/
static int load_map_line(char *line, int width, int line_num)
{
  int i;

  for (i = 0; i < width; i++)
    {
      switch (line[i])
        {
        case '.':
          game_map[line_num][i+1].type = TERRAIN_PLAIN;
          break;

        case '~':
          game_map[line_num][i+1].type = TERRAIN_WATER;
          break;

        case '#':
          game_map[line_num][i+1].type = TERRAIN_WALL;
          break;

        case '@':
          game_map[line_num][i+1].type = TERRAIN_HOME;
          break;

        default:
          fprintf(stderr, "parse_map_line(): Unknown Type: %c\n", line[i]);
          return(ERROR);
        }
    }

  return(OK);
}


/***************************************************
 * process_map()                                   *
 *                                                 *
 ***************************************************/
static int process_map(void)
{
  int i;
  int j;
  int width;
  int height;

  width  = game_state.map_size_x;
  height = game_state.map_size_y;

  /* Set up the links on the map A-star pathfinding
   *
   * NOTE: South is at 0, North is at (height-1)
   */
  for (i = 1; i <= (height-2); i++)
    {
      for (j = 1; j <= (width-2); j++)
        {
          /* North
           */
          if ((game_map[i][j].type == TERRAIN_PLAIN) |
              (game_map[i][j].type == TERRAIN_HOME))
            {
              if ((game_map[i+1][j].type == TERRAIN_WALL) |
                  (game_map[i+1][j].type == TERRAIN_WATER))
                {
                  game_map[i][j].links &= ~LINK_NORTH;
                }
              else
                {
                  game_map[i][j].links |=  LINK_NORTH;
                }
              
              /* South
               */
              if ((game_map[i-1][j].type == TERRAIN_WALL) |
                  (game_map[i-1][j].type == TERRAIN_WATER))
                {
                  game_map[i][j].links &= ~LINK_SOUTH;
                }
              else
                {
                  game_map[i][j].links |=  LINK_SOUTH;
                }
              
              /* East
               */
              if ((game_map[i][j+1].type == TERRAIN_WALL) |
                  (game_map[i][j+1].type == TERRAIN_WATER))
                {
                  game_map[i][j].links &= ~LINK_EAST;
                }
              else
                {
                  game_map[i][j].links |=  LINK_EAST;
                }
              
              /* West
               */
              if ((game_map[i][j-1].type == TERRAIN_WALL) |
                  (game_map[i][j-1].type == TERRAIN_WATER))
                {
                  game_map[i][j].links &= ~LINK_WEST;
                }
              else
                {
                  game_map[i][j].links |=  LINK_WEST;
                }
            }
          else
            {
              /* Can't move onto a Wall or Water, so links
               * off of it don't make sense.
               */
              game_map[i][j].links = 0x0;
            }
        }
    }

  /* Handle Edges
   */
  for (j = 0; j < width; j++)
    {
      /* North Edge
       */
      game_map[height-1][j].links = 0x0;

      /* South Edge
       */
      game_map[0][j].links = 0x0;
    }

  for (i = 0; i < height; i++)
    {
      /* East Edge
       */
      game_map[i][width-1].links = 0x0;

      /* West Edge
       */
      game_map[i][0].links = 0x0;
    }

  /* Load Node ID Numbers & Zero Data
   */
  for (i = 0; i < height; i++)
    {
      for (j = 0; j < width; j++)
        {
          game_map[i][j].x = j;
          game_map[i][j].y = i;

          game_map[i][j].g = 0;
          game_map[i][j].h = 0;
          game_map[i][j].f = 0;
          game_map[i][j].parent = NULL;

          game_map[i][j].visited  = 0;
          game_map[i][j].packages = 0;
        }
    }

  return(OK);
}


/***************************************************
 * print_map()                                     *
 *                                                 *
 ***************************************************/
static void print_map(void)
{
  int i;
  int j;
  int width;
  int height;

  width  = game_state.map_size_x;
  height = game_state.map_size_y;

  for (i = height-1; i >= 0; i--)
    {
      for (j = 0; j < width; j++)
        {
          switch (game_map[i][j].type)
            {
            case TERRAIN_PLAIN:
              printf(".");
              break;

            case TERRAIN_WATER:
              printf("~");
              break;

            case TERRAIN_WALL:
              printf("#");
              break;

            case TERRAIN_HOME:
              printf("@");
              break;

            default:
              printf("!");
              break;
            }
        }
      printf("\n");
    }
}


/***************************************************
 * print_map_packages()                            *
 *                                                 *
 ***************************************************/
static void print_map_packages(void)
{
  int i;
  int j;
  int width;
  int height;

  width  = game_state.map_size_x;
  height = game_state.map_size_y;

  for (i = height-1; i >= 0; i--)
    {
      for (j = 0; j < width; j++)
        {
          if (game_map[i][j].packages > 9)
            {
              printf("9");
            }
          else
            {
              printf("%d", game_map[i][j].packages);
            }
        }
      printf("\n");
    }
}


/***************************************************
 * print_map_links()                               *
 *                                                 *
 ***************************************************/
static void print_map_links(void)
{
  int i;
  int j;

  int width;
  int height;

  width  = game_state.map_size_x;
  height = game_state.map_size_y;

  for (i = height-1; i >= 0; i--)
    {
      for (j = 0; j < width; j++)
        {
          if (game_map[i][j].links & LINK_NORTH)
            {
              printf(" | ");
            }
          else
            {
              printf(" . ");
            }
        }
      printf("\n");

      for (j = 0; j < width; j++)
        {
          if (game_map[i][j].links & LINK_WEST)
            {
              printf("_*");
            }
          else
            {
              printf(".*");
            }
          if (game_map[i][j].links & LINK_EAST)
            {
              printf("_");
            }
          else
            {
              printf(".");
            }
        }
      printf("\n");

      for (j = 0; j < width; j++)
        {
          if (game_map[i][j].links & LINK_SOUTH)
            {
              printf(" | ");
            }
          else
            {
              printf(" . ");
            }
        }
      printf("\n");

    } /* Loop over rows
       */
}


/***************************************************
 * print_path()                                    *
 *                                                 *
 ***************************************************/
#ifdef DEBUG_PF
void print_path(Path_entry *head)
{
  Path_entry *tmp;

  tmp = head;

  printf("Path:\n");
  while (tmp != NULL)
    {
      printf("x: %d y: %d\n", tmp->node->x, tmp->node->y);
      tmp = tmp->next;
    }
  printf("End of Path.\n");
}
#endif


/***************************************************
 * load_cfg()                                      *
 *                                                 *
 ***************************************************/
static int load_cfg(void)
{
  char   buffer[80];
  char  *pbuffer;
  char **stringp;
  char  *token;
  int    status;

  int    robot_id;
  int    max_weight;
  int    money;

  /* The configuration consists of:
   *
   * 1) A unique identifier,
   * 2) the maximum carrying capacity,
   * 3) the initial amount of money (used for bidding)
   *
   * These are all integers and are sent on one line,
   * separated by spaces, in the above order.
   */
  status = socket_fgets(buffer, 80);
  if (status != OK)
    {
      fprintf(stderr, "load_cfg(): Error in socket_fgets().\n");
      return(ERROR);
    }

  pbuffer    = &(buffer[0]);
  stringp    = &(pbuffer);

  token      = strsep(stringp, " ");
  if (token == NULL)
    {
      fprintf(stderr, "load_cfg(): Unexpected NULL token.\n");
      fprintf(stderr, "line: __LINE__\n");
      return(ERROR);
    }
  robot_id   = atoi(token);

  token      = strsep(stringp, " ");
  if (token == NULL)
    {
      fprintf(stderr, "load_cfg(): Unexpected NULL token.\n");
      fprintf(stderr, "line: __LINE__\n");
      return(ERROR);
    }
  max_weight = atoi(token);

  token      = strsep(stringp, " ");
  if (token == NULL)
    {
      fprintf(stderr, "load_cfg(): Unexpected NULL token.\n");
      fprintf(stderr, "line: __LINE__\n");
      return(ERROR);
    }
  money      = atoi(token);

  printf("init_cfg(): ID: %d Weight: %d Money: %d\n",
         robot_id, max_weight, money);

  /* Update Game State
   */
  game_state.robot_id   = robot_id;
  game_state.max_weight = max_weight;
  game_state.money      = money;

  return(OK);
}


/***************************************************
 * process_server_resp()                           *
 *                                                 *
 ***************************************************/
static int process_server_resp(void)
{
  static char func[] = "process_server_resp()";

  char  *buffer;
  char  *pbuffer;
  char **stringp;
  char  *token;
  int    status;
  int    read_token;

  int    robot_id;
  int    robot_x;
  int    robot_y;
  int    package_id;
  int    dest_x;
  int    dest_y;
  int    weight;

  int    i;

  int    num_robot_moves;
  int    replan_path;
  
  /* Server's reply to commands
   *
   * At the end of each turn the server will send a reply showing what
   * happened to all robots during the turn. All the players will
   * receive the same information.
   *
   * The format of the response is a list of robot updates.  Each
   * robot update is of the form #robot_id and then the actions in the
   * order they happened.
   *
   * A robot is listed if and only if it was alive at the beginning of
   * the turn. The following actions are possible:
   *
   * N The robot moved north.
   * S The robot moved south.
   * E The robot moved east.
   * W The robot moved west.
   * P id The robot picked up package with id id.
   * D id The robot dropped package with id id.
   * X num1 Y num2 The robots location is set to (num1,num2).
   */
  status = socket_fgets_unlim(&buffer);
  if (status != OK)
    {
      fprintf(stderr, "%s: Error in socket_fgets_unlim().\n", func);
      free(buffer);
      return(ERROR);
    }

  printf("resp: %s", buffer);


  /****************************************
   * Loop through the string and parse it *
   ****************************************/
  pbuffer     = &(buffer[0]);
  stringp     = &(pbuffer);
  read_token  = TRUE;
  token       = NULL;
  replan_path = FALSE;

  while (TRUE)
    {
      if (read_token)
        {
          token = strsep(stringp, " ");
        }
      
      if (token == NULL)
        {
          break;
        }
     
      switch (token[0])
        {
        case '#':
          /* Found the # character. Skip past it.
           */
          token++;
          robot_id = atoi(token);
          printf("Robot ID: %d\n", robot_id);

          /* If a robot moves more than 1 time per round, it was pushed.
           */
          num_robot_moves = 0;
          break;
          
        case 'N':
          num_robot_moves++;
          if (robot_id == game_state.robot_id)
            {
              game_state.robot_y++;
              if (num_robot_moves > 1)
                {
                  printf("!!**!! Robot %d Pushed N (%d, %d)\n",
                         robot_id, game_state.robot_x, game_state.robot_y);
                  replan_path = TRUE;
                }
              else
                {
                  printf("!! Robot %d Moved  N (%d, %d)\n",
                         robot_id, game_state.robot_x, game_state.robot_y);
                }                  
            }
          else
            {
              if (num_robot_moves > 1)
                {
                  printf("Robot %d Pushed N\n", robot_id);
                }
              else
                {
                  printf("Robot %d Moved  N\n", robot_id);
                }
            }
          break;
          
        case 'S':
          num_robot_moves++;
          if (robot_id == game_state.robot_id)
            {
              game_state.robot_y--;
              if (num_robot_moves > 1)
                {
                  printf("!!**!! Robot %d Pushed S (%d, %d)\n",
                         robot_id, game_state.robot_x, game_state.robot_y);
                  replan_path = TRUE;
                }
              else
                {
                  printf("!! Robot %d Moved  S (%d, %d)\n",
                         robot_id, game_state.robot_x, game_state.robot_y);
                }
            }
          else
            {
              if (num_robot_moves > 1)
                {
                  printf("Robot %d Pushed S\n", robot_id);
                }
              else
                {
                  printf("Robot %d Moved  S\n", robot_id);
                }
            }
          break;
          
        case 'E':
          num_robot_moves++;
          if (robot_id == game_state.robot_id)
            {
              game_state.robot_x++;
              if (num_robot_moves > 1)
                {
                  printf("!!**!! Robot %d Pushed E (%d, %d)\n",
                         robot_id, game_state.robot_x, game_state.robot_y);
                  replan_path = TRUE;
                }
              else
                {
                  printf("!! Robot %d Moved  E (%d, %d)\n",
                         robot_id, game_state.robot_x, game_state.robot_y);
                }
            }
          else
            {
              if (num_robot_moves > 1)
                {
                  printf("Robot %d Pushed E\n", robot_id);
                }
              else
                {
                  printf("Robot %d Moved  E\n", robot_id);
                }
            }
          break;
          
        case 'W':
          num_robot_moves++;
          if (robot_id == game_state.robot_id)
            {
              game_state.robot_x--;
              if (num_robot_moves > 1)
                {
                  printf("!!**!! Robot %d Pushed W (%d, %d)\n",
                         robot_id, game_state.robot_x, game_state.robot_y);
                  replan_path = TRUE;
                }
              else
                {
                  printf("!! Robot %d Moved  W (%d, %d)\n",
                         robot_id, game_state.robot_x, game_state.robot_y);
                }
            }
          else
            {
              if (num_robot_moves > 1)
                {
                  printf("Robot %d Pushed W\n", robot_id);
                }
              else
                {
                  printf("Robot %d Moved  W\n", robot_id);
                }
            }
          break;
          
        case 'P':
#ifdef DEBUG_RESP
          printf("Robot %d Picked\n", robot_id);
#endif
          token = strsep(stringp, " ");
          if (token == NULL)
            {
              printf("Robot %d Picked No Package\n", robot_id);
              read_token = FALSE; /* Skip a read */
              break;
            }

          /* Picklist can be empty
           */
          if (!isdigit(token[0]))
            {
              printf("Robot %d Picked No Package\n", robot_id);
              read_token = FALSE; /* Skip a read */
            }
          else
            {
              package_id = atoi(token);
              if (robot_id == game_state.robot_id)
                {
                  printf("!! Robot %d Picked Package %d\n",
                         robot_id, package_id);

                  status = package_get_info(game_state.pkg_unpicked,
                                            package_id, &dest_x,
                                            &dest_y,    &weight);
                  if (status != OK)
                    {
                      fprintf(stderr,
                              "process_server_resp(): Error line: %d\n",
                              __LINE__);
                      return(ERROR);
                    }
                  
                  status = package_add_entry(&(game_state.pkg_picked),
                                             package_id, dest_x,
                                             dest_y,     weight);
                  if (status != OK)
                    {
                      fprintf(stderr,
                              "process_server_resp(): Error line: %d\n",
                              __LINE__);
                      return(ERROR);
                    }
                  
                  /* Update the amount of weight we're carrying
                   */
                  game_state.weight += weight;
                }
              else
                {
                  printf("Robot %d Picked Package %d\n",
                         robot_id, package_id);                  
                }

            } /* Package Picked */

          break;
          
        case 'D':
#ifdef DEBUG_RESP
          printf("Robot %d Dropped\n", robot_id);
#endif
          token = strsep(stringp, " ");
          if (token == NULL)
            {
              printf("Robot %d Dropped No Package\n", robot_id);
              read_token = FALSE; /* Skip a read */
              break;
            }

          /* Droplist can be empty
           */
          if (!isdigit(token[0]))
            {
              printf("Robot %d Dropped No Package\n", robot_id);
              read_token = FALSE; /* Skip a read */
            }
          else
            {
              package_id = atoi(token);
              if (robot_id == game_state.robot_id)
                {
                  printf("!! Robot %d Dropped Package %d at %d, %d\n",
                         robot_id, package_id,
                         game_state.robot_x,
                         game_state.robot_y);

                  status = package_remove_entry(&(game_state.pkg_picked),
                                package_id, &dest_x, &dest_y, &weight);
                  if (status != OK)
                    {
                      fprintf(stderr,
                              "process_server_resp(): Error line: %d\n",
                              __LINE__);
                      return(ERROR);
                    }

#ifdef DEBUG_PKG
                  printf("Removed package: %d from list with wgt: %d\n",
                         package_id, weight);

                  print_package(game_state.pkg_picked);
#endif

                  if ((dest_x != game_state.robot_x) ||
                      (dest_y != game_state.robot_y))
                    {
                      printf("!!**!! Package Dropped. Robot Pushed!\n");
                      replan_path = TRUE;
                    }
                  else
                    {
                      /* Update the current guess of our score
                       *
                       * If it was dropped, we probably didn't
                       * get any credit for it.
                       */
                      game_state.score_guess += weight;
                    }
                   
                  /* Update the amount of weight we're carrying
                   */
                  game_state.weight -= weight;
                }
              else
                {
                  printf("Robot %d Dropped Package %d\n",
                         robot_id, package_id);                  
                }

            } /* Package Dropped */
          break;
          
        case 'X':
          token = strsep(stringp, " ");
          if (token == NULL)
            {
              printf("Robot %d Set X Position, but No Location Given!\n",
                     robot_id);
              read_token = FALSE;
              break;
            }
          robot_x = atoi(token);
          if (robot_id == game_state.robot_id)
            {
              printf("!! Robot %d Set X Position: %d\n", robot_id, robot_x);
              game_state.robot_x = robot_x;
            }
          else
            {
              printf("Robot %d Set X Position: %d\n", robot_id, robot_x);
            }
          break;
          
        case 'Y':
          token = strsep(stringp, " ");
          if (token == NULL)
            {
              printf("Robot %d Set Y Position, but No Location Given!\n",
                     robot_id);
              read_token = FALSE;
              break;
            }
          robot_y = atoi(token);
          if (robot_id == game_state.robot_id)
            {
              printf("!! Robot %d Set Y Position: %d\n", robot_id, robot_y);
              game_state.robot_y = robot_y;
            }
          else
            {
              printf("Robot %d Set Y Position: %d\n", robot_id, robot_y);
            }
          break;
          
        case 'R':
          /* New Death Message from the Server
           */
          printf("%s %s", token, *stringp);

          /* Handle death message
           */
          free(buffer);

          for (i = 0; i < 3; i++)
            {
              status = socket_fgets_unlim(&buffer);
              if (status != OK)
                {
                  fprintf(stderr, "%s: Error line: %d.\n", func, __LINE__);
                  free(buffer);
                  return(ERROR);
                }
              printf("%s", buffer);
              free(buffer);
            }

          game_state.alive = FALSE;
          return(OK);

          break;

        default:
          printf("Unknown Character: %c\n", token[0]);
          break;
        }
    }

  /* Free allocated memory
   */
  free(buffer);

#ifdef DEBUG_PKG
  print_package(game_state.pkg_picked);
#endif

  game_state.replan_path = replan_path;

  return(OK);
}


/***************************************************
 * game_play()                                     *
 *                                                 *
 ***************************************************/
static int game_play(void)
{
  int status;

  /* The game is made up of a number of turns. A turn looks as follows:
   *
   * 1) Server: a list of packages available at the player's current location
   * 2) Client: issue a command to the robot
   * 3) Server: server responds with what happened during the turn
   */

  while (game_state.alive)
    {
      printf("Score (Guess): %d  Money: %d\n",
             game_state.score_guess, game_state.money);

      status = get_package_list();
      if (status != OK)
        {
          fprintf(stderr, "game_play(): Error in get_package_list().\n");
          return(ERROR);
        }

      if (game_state.alive == 0)
        {
          /* Getting reply in the package comm for some reason.
           */
          break;
        }

      status = issue_command();
      if (status != OK)
        {
          fprintf(stderr, "game_play(): Error in issue_command().\n");
          return(ERROR);
        }

      status = process_server_resp();
      if (status != OK)
        {
          fprintf(stderr, "game_play(): Error in process_server_resp().\n");
          return(ERROR);
        }

      printf("\n");
    }

  return(OK);
}


/***************************************************
 * get_package_list()                              *
 *                                                 *
 ***************************************************/
static int get_package_list(void)
{
  char  *buffer;
  char  *pbuffer;
  char **stringp;
  char  *token;
  int    status;

  int            p_id;
  int            p_dest_x;
  int            p_dest_y;
  int            p_weight;
  Package_entry *p_list;
  int            i;
  int            num_pkg;

  int            x;
  int            y;

  printf("Looking for Packages...\n");

  /* Clear any previous list of packages
   */
  package_delete(&(game_state.pkg_unpicked));

  /* Get the new list of packages.
   */
  status = socket_fgets_unlim(&buffer);
  if (status != OK)
    {
      fprintf(stderr, "get_package_list(): Error in socket_fgets_unlim().\n");
      free(buffer);
      return(ERROR);
    }

  /* When Game Over happens, the response gets printed out as
   * part of the package list, rather than as part of the
   * server response ... ugh. Need to capture that here too.
   */
  if (buffer[0] == 'R')
    {
      printf("%s", buffer);
      free(buffer);

      for (i = 0; i < 3; i++)
        {
          status = socket_fgets_unlim(&buffer);
          if (status != OK)
            {
              fprintf(stderr,
                      "get_package_list(): Error line: %d.\n", __LINE__);
              free(buffer);
              return(ERROR);
            }
          printf("%s", buffer);
          free(buffer);
        }
      
      game_state.alive = FALSE;
      return(OK);
    }

  if (strlen(buffer) == 1)
    {
      printf("No packages at current location.\n");

      /* Update the game state
       */
      x = game_state.robot_x;
      y = game_state.robot_y;
      game_map[y][x].packages = 0;

      free(buffer);
      return(OK);
    }

  /* Parse package list
   */
  pbuffer    = &(buffer[0]);
  stringp    = &(pbuffer);
  token      = NULL;

  p_list     = NULL;
  num_pkg    = 0;

  printf("I See Packages!:\n");
  while (TRUE)
    {
      /* Get Package ID
       */
      token = strsep(stringp, " ");
      if (token == NULL)
        {
          /* No more packages left, bail out
           */
          break;
        }
      p_id = atoi(token);


      /* Get Package Destination X
       */
      token = strsep(stringp, " ");
      if (token == NULL)
        {
          fprintf(stderr, "get_package_list(): Unexpected NULL token.\n");
          break;
        }
      p_dest_x = atoi(token);


      /* Get Package Destination Y
       */
      token = strsep(stringp, " ");
      if (token == NULL)
        {
          fprintf(stderr, "get_package_list(): Unexpected NULL token.\n");
          break;
        }
      p_dest_y = atoi(token);


      /* Get Package Weight
       */
      token = strsep(stringp, " ");
      if (token == NULL)
        {
          fprintf(stderr, "get_package_list(): Unexpected NULL token.\n");
          break;
        }
      p_weight = atoi(token);


      /* Print the package
       */
      printf("package: ID: %d x: %d y: %d wgt: %d\n",
             p_id, p_dest_x, p_dest_y, p_weight);

      /* Add the package to the list of unpicked packages
       */
      status = package_add_entry(&p_list, p_id, p_dest_x, p_dest_y, p_weight);
      if (status != OK)
        {
          fprintf(stderr, "get_package_list(): Error line: %d\n", __LINE__);
          package_delete(&p_list);
          free(buffer);
          return(ERROR);
        }

      num_pkg++;
    }
  printf("End Package List.\n");

  /* Update the game state
   */
  game_state.pkg_unpicked = p_list;

  /* Store the number of packages at this location
   */
  x = game_state.robot_x;
  y = game_state.robot_y;
  if (num_pkg > 255)
    {
      num_pkg = 255; /* Only store 1 byte per node */
    }
  game_map[y][x].packages = num_pkg;

#ifdef DEBUG_PKG2
  print_package(game_state.pkg_unpicked);
#endif

  /* Free allocated memory
   */
  free(buffer);
  return(OK);
}


/**************************************************
 * package_add_entry()                            *
 *                                                *
 **************************************************/
static int package_add_entry(Package_entry **head,
                             int id, int dest_x, int dest_y, int weight)
{
  Package_entry *new_entry;
  Package_entry *tmp;

  new_entry = (Package_entry *) malloc(sizeof(Package_entry));
  if (new_entry == NULL)
    {
      fprintf(stderr, "package_add_entry(): Error in malloc().\n");
      return(ERROR);
    }

  new_entry->id     = id;
  new_entry->dest_x = dest_x;
  new_entry->dest_y = dest_y;
  new_entry->weight = weight;

  new_entry->next   = NULL;
  new_entry->prev   = NULL;

  if (*head == NULL)
    {
      *head = new_entry;
    }
  else
    {
#define DEPENDABLE
#ifdef  DEPENDABLE
      tmp = *head;
      while (tmp->next != NULL)
        {
          tmp = tmp->next;
        }

      tmp->next = new_entry;
      new_entry->prev = tmp;
#else
      /* Ick this is nasty. It's really late ...
       * Warning! This is what happens when you code without much sleep!
       * Trying to sort so that highest weight is at the front.
       */
      tmp = *head;
      while (tmp->next != NULL)
        {
          if (tmp->next->weight < weight)
            {
              break;
            }

          tmp = tmp->next;
        }

      if ((tmp->next == NULL) || (tmp->prev == NULL))
        {
          /* Special Cases
           */
          if (tmp->prev == NULL)
            {
              /* We're at the head of the list
               */
              if (weight > tmp->weight)
                {
                  /* Add as head
                   */
                  new_entry->next = *head;
                  (*head)->prev   =  new_entry;
                  *head           =  new_entry;
                }
              else
                {
                  /* Insert after head
                   */
                  new_entry->next = (*head)->next;
                  new_entry->prev = (*head);
                  (*head)->next   = new_entry;
                }
            }
          else /* We're at the end of the list */
            {
              tmp->next = new_entry;
              new_entry->prev = tmp;
            }
        }
      else
        {
          next = tmp->next;

          /* General Case
           */
          new_entry->prev = tmp;
          new_entry->next = tmp->next;

          tmp->next       = new_entry;
          tmp->next->prev = new_entry;
        }
#endif /* DEPENDABLE */
    }

#ifdef DEBUG_PKG
  printf("package_add_entry(): Added: %d %d %d %d\n",
         id, dest_x, dest_y, weight);
#endif

  return(OK);
}


/**************************************************
 * print_package()                                *
 *                                                *
 **************************************************/
#ifdef DEBUG_PKG2
static void print_package(Package_entry *head)
{
  Package_entry *tmp;

  printf("Package List!\n");

  tmp = head;

  while (tmp != NULL)
    {
      printf("package: %d %d %d %d\n",
             tmp->id, tmp->dest_x, tmp->dest_y, tmp->weight);

      tmp = tmp->next;
    }

  printf("End Package List.\n");
}
#endif

/**************************************************
 * package_get_info()                             *
 *                                                *
 **************************************************/
static int package_get_info(Package_entry *head, int id,
                            int *dest_x, int *dest_y, int *weight)
{
  Package_entry *tmp;

  if (head == NULL)
    {
      fprintf(stderr, "package_get_info(): Package list empty!\n");
      return(ERROR);
    }

  tmp = head;

  while (tmp != NULL)
    {
      if (tmp->id == id)
        {
          /* Found!
           */
          *dest_x = tmp->dest_x;
          *dest_y = tmp->dest_y;
          *weight = tmp->weight;

#ifdef DEBUG_PKG
          printf("package_get_info(): Got ID: %d X: %d Y: %d Wgt: %d\n",
                 id, *dest_x, *dest_y, *weight);
#endif
          return(OK);
        }

      tmp = tmp->next;
    }

  /* Not found
   */
  return(ERROR);
}


/**************************************************
 * package_remove_entry()                         *
 *                                                *
 **************************************************/
static int package_remove_entry(Package_entry **head,
                                int id, int *x, int *y, int *wgt)
{
  Package_entry *tmp;
  Package_entry *prev;
  Package_entry *next;

  if (head == NULL)
    {
      fprintf(stderr, "package_remove_entry(): Package list empty!\n");
      return(ERROR);
    }

  tmp = *head;

  while (tmp != NULL)
    {
      if (tmp->id == id)
        {
          /* Found!
           */
          *x   = tmp->dest_x;
          *y   = tmp->dest_y;
          *wgt = tmp->weight;

          /* Remove the entry
           */
          if ((tmp->prev == NULL) || (tmp->next == NULL))
            {
              /* Special Cases
               */
              if (tmp->prev == NULL)
                {
                  /* Removed node is the head of the list
                   */
                  *head = tmp->next;
                  if (*head != NULL)
                    {
                      /* i.e. More than 1 entry in pque
                       */
                      (*head)->prev = NULL;
                    }
                }
              else
                {
                  /* Removed node is the tail of the pque
                   */
                  (tmp->prev)->next = NULL;
                }
            }
          else
            {
              /* General Case
               */
              prev = tmp->prev;
              next = tmp->next;
              
              prev->next = next;
              next->prev = prev;
            }

          free(tmp);
          return(OK);
        }

      tmp = tmp->next;
    }

  /* Not found
   */
  return(ERROR);
}






/**************************************************
 * package_delete()                               *
 *                                                *
 **************************************************/
static void package_delete(Package_entry **head)
{
  Package_entry *tmp;

  while (*head != NULL)
    {
      tmp = (*head)->next;

#ifdef DEBUG_PKG
      printf("Deleting package: %d %d %d %d\n",
             (*head)->id, (*head)->dest_x, (*head)->dest_y, (*head)->weight);
#endif

      free(*head);

      *head = tmp;
    }
}


/**************************************************
 * issue_command()                                *
 *                                                *
 **************************************************/
static int issue_command(void)
{
  char command[1024];
  char command_tmp[1024];
  int  status;
  int  robot_x;
  int  robot_y;
  int  x;
  int  y;
  int  moved;
  int  cmd;
  Path_entry    *path;
  Map_node      *node;
  Package_entry *pkg;
  int            wgt_avail;
  int            picked_package;
  int            dropped_package;

  /*******************************
   * Determine what to do.       *
   *******************************/

  /* Get our current location
   */
  robot_x = game_state.robot_x;
  robot_y = game_state.robot_y;


  /* If we were pushed off the path we were following,
   * let's just blow the currrent path away and replan
   * for the current package on the list. (Which is
   * usually the next, since we drop a package).
   */
  if ((game_state.active_path != NULL) &&
      (game_state.replan_path == TRUE))
    {
      printf("!!**!! Pushed Off the Path!! Replanning path.\n");
      path_delete(&(game_state.active_path));
    }


  /* First, are we sitting in a square that has packages?
   * If so, try to pick them up.
   */
  if (game_state.pkg_unpicked != NULL)
    {
      /* Go through the list and request to pick up as many
       * packages as we can.
       *
       * We might want to pick the heaviest packages first
       * here, so that we can maximize our score.
       *
       * To do this, we put them into the list in weight order.
       *
       * In Addition, we can produce a weighted score for the
       * package, which includes its weight and the distance
       * we have to carry it. However to do this we have to
       * compute the path to travel to each package's destination.
       */
      wgt_avail = (game_state.max_weight - game_state.weight);
      picked_package = FALSE;

      pkg = game_state.pkg_unpicked;

      printf("max_weight: %d\n", game_state.max_weight);
      printf("weight    : %d\n", game_state.weight);
      printf("wgt_avail : %d\n", wgt_avail);

      while ((pkg != NULL) && (wgt_avail >= 0))
        {
          /* Only pick up packages that don't put us over weight.
           * We don't need to worry about picking up packages
           * that have already been delivered, since they disappear.
           */
          if (pkg->weight <= wgt_avail)
            {
              printf("Requesting to pick package: %d %d %d %d\n",
                     pkg->id, pkg->dest_x, pkg->dest_y, pkg->weight);

              if (!picked_package)
                {
                  sprintf(command, "1 Pick %d", pkg->id);
                  picked_package = TRUE;
                }
              else
                {
                  sprintf(command_tmp, " %d", pkg->id);
                  strcat(command, command_tmp);
                }
              wgt_avail -= pkg->weight;
            }

          pkg = pkg->next;
        }

      /* This info returned by the server again next round.
       * It's stale, so we should delete it.
       *
       * Unfortunately, we still need it, when we process the server
       * response, in order to update the game state for the weights
       * and destinations of what we're carrying.
       * 
       * (When the server tells us we've picked something,
       *  only the ID is sent.)
       */
      if (picked_package)
        {
          strcat(command, "\n");

          printf("command: %s", command);

          status = socket_write(command, strlen(command));
          if (status != OK)
            {
              fprintf(stderr, "issue_command(): Error in socket_write().\n");
              return(ERROR);
            }

          /* Done for this round */
          game_state.money--;
          return(OK);
        }
      else
        {
          printf("Can't pick up any more packages. :(\n");
          printf("Let's continue to follow the path, etc.\n");
        }
    }  


  /* Do we have an active path that
   * we're already following?
   */
  if (game_state.active_path != NULL)
    {
      printf("issue_command(): Following path.\n");

      /* Follow the path
       */
      status = follow_path(robot_x, robot_y,
                           &(game_state.active_path), command);
      if (status != OK)
        {
          fprintf(stderr, "issue_command(): Error in follow_path().\n");
          return(ERROR);
        }

      printf("command: %s", command);
      status = socket_write(command, strlen(command));
      if (status != OK)
        {
          fprintf(stderr, "issue_command(): Error in socket_write().\n");
          return(ERROR);
        }

      /* Done for this round
       */
      game_state.money--;
      return(OK);
    }

  
  /* If not, look at our list of packages.
   * Do we have any packages at the moment?
   */
  if (game_state.pkg_picked != NULL)
    {
      /* We have a package. Did we just arrive at the destination?
       * If so drop the package. Package will be removed from the
       * pkg_picked list when it's confirmed dropped by the server.
       */
      pkg = game_state.pkg_picked;
      dropped_package = FALSE;
      while (pkg != NULL)
        {
          /* Drop all the packages that have this address.
           */
          if ((pkg->dest_x == robot_x) && (pkg->dest_y == robot_y))
            {
              if (!dropped_package)
                {
                  printf("We've arrived! Let's drop off some packages.\n");
                }

              printf("Requesting to drop package: %d %d %d %d\n",
                     pkg->id, pkg->dest_x, pkg->dest_y, pkg->weight);

              if (!dropped_package)
                {
                  sprintf(command, "1 Drop %d", pkg->id);
                  dropped_package = TRUE;
                }
              else
                {
                  sprintf(command_tmp, " %d", pkg->id);
                  strcat(command, command_tmp);
                }
            }

          pkg = pkg->next;

        } /* Loop over pkg_picked */

      if (dropped_package)
        {
          strcat(command, "\n");
          printf("command: %s", command);
        }

      /* If we're don't have anything to drop off here, let's
       * plan a path to deliver the first package on the list.
       */
      if (!dropped_package)
        {
          /* Otherwise, plan a path to deliver it.
           */
          x = game_state.pkg_picked->dest_x;
          y = game_state.pkg_picked->dest_y;

          printf("Planning path to deliver a package!\n");
          printf("Planning Path from (%d, %d) to (%d, %d)\n",
             robot_x, robot_y, x, y);

          path   = NULL;
          status = plan_path(robot_x, robot_y, x, y, &path);
          if (status != OK)
            {
              fprintf(stderr, "issue_command(): Error in plan_path().\n");
              path_delete(&path);
              return(ERROR);
            }

#ifdef DEBUG_PF
          print_path(path);
#endif
          
          /* Update the game state
           */
          game_state.active_path = path;
          
          /* First node is always our start position (Remove it).
           */
          status = path_pop_node(&(game_state.active_path), &node);
          if (status != OK)
            {
              fprintf(stderr, "follow_path(): Error in path_pop_node().\n");
              return(ERROR);
            }
          
          printf("Popped path starting loc: x %d y %d\n", node->x, node->y);
          
          if (path_length(game_state.active_path) >= 1)
            {
              /* Follow the path
               */
              status = follow_path(robot_x, robot_y,
                                   &(game_state.active_path), command);
              if (status != OK)
                {
                  fprintf(stderr,
                          "issue_command(): Error in follow_path().\n");
                  return(ERROR);
                }
            }
          else
            {
              /* We're sitting on our destination already. No-Op.
               */
              sprintf(command, "1 Drop %d\n", game_state.pkg_picked->id);
            }

        } /* Planned Path */

      printf("command: %s", command);
      status = socket_write(command, strlen(command));
      if (status != OK)
        {
          fprintf(stderr, "issue_command(): Error in socket_write().\n");
          return(ERROR);
        }

      /* Done for this round
       */
      game_state.money--;
      return(OK);
    }


  /* Ok :
   *
   * 1) We're don't have any packages
   * 2) We're not sitting on any
   * 3) We're not moving along a path already
   * 
   * So ...
   *
   * Let's head for a home base to find something to do.
   */
  status = find_closest_home(&x, &y);
  if (status != OK)
    {
      printf("issue_command(): Sadness, no home base found.\n");
      printf("issue_command(): Beginning Aimless Wandering.\n");

      moved = FALSE;
      while (!moved)
        {
          cmd = (int) (4.0 * drand48());
          if (game_map[robot_y][robot_x].links & (1 << cmd))
            {
              switch (cmd)
                {
                case 0:
                  sprintf(command, "1 Move N\n");
                  break;
                case 1:
                  sprintf(command, "1 Move S\n");
                  break;
                case 2:
                  sprintf(command, "1 Move E\n");
                  break;
                case 3:
                  sprintf(command, "1 Move W\n");
                  break;
                }
              moved = TRUE;
            }
        }
    }
  else
    {
      printf("Alas, nothing to do here. Let's look for a home base.\n");
      printf("Found closest & least visited home base at: (%d, %d)\n", x, y);
      printf("Planning Path from (%d, %d) to (%d, %d)\n",
             robot_x, robot_y, x, y);

      path = NULL;
      status = plan_path(robot_x, robot_y, x, y, &path);
      if (status != OK)
        {
          fprintf(stderr, "issue_command(): Error in plan_path().\n");
          path_delete(&path);
          return(ERROR);
        }

#ifdef DEBUG_PF
      print_path(path);
#endif

      /* Update the game state
       */
      game_state.active_path = path;

      /* First node is always our start position (Remove it).
       */
      status = path_pop_node(&(game_state.active_path), &node);
      if (status != OK)
        {
          fprintf(stderr, "follow_path(): Error in path_pop_node().\n");
          return(ERROR);
        }

#ifdef DEBUG_PF
      printf("Popped path starting loc: x %d y %d\n", node->x, node->y);
#endif

      if (path_length(game_state.active_path) >= 1)
        {
          /* Follow the path
           */
          status = follow_path(robot_x, robot_y,
                               &(game_state.active_path), command);
          if (status != OK)
            {
              fprintf(stderr, "issue_command(): Error in follow_path().\n");
              return(ERROR);
            }
        }
      else
        {
          /* We're sitting on our destination already. No-Op.
           */
          sprintf(command, "1 Drop\n");
        }
    }

  printf("command: %s", command);
  status = socket_write(command, strlen(command));
  if (status != OK)
    {
      fprintf(stderr, "issue_command(): Error in socket_write().\n");
      return(ERROR);
    }

  game_state.money--;
  return(OK);
}


/***************************************************
 * follow_path()                                   *
 *                                                 *
 ***************************************************/
static int follow_path(int robot_x, int robot_y, Path_entry **path, char *cmd)
{
  Map_node *node;
  int       status;
  int       new_x;
  int       new_y;
  int       diff_x;
  int       diff_y;
  int       dest_x;
  int       dest_y;

  Path_entry *tmp;

  if (*path == NULL)
    {
      fprintf(stderr, "follow_path(): Path has no nodes.\n");
      return(ERROR);
    }
   
  status = path_pop_node(path, &node);
  if (status != OK)
    {
      fprintf(stderr, "follow_path(): Error in path_pop_node().\n");
      return(ERROR);
    }

  new_x = node->x;
  new_y = node->y;

#ifdef  DEBUG_FOLLOW_PATH
  printf("follow_path(): current x: %d y: %d, New x: %d y: %d\n",
         robot_x, robot_y, new_x, new_y);
#endif

  diff_x = new_x - robot_x;
  diff_y = new_y - robot_y;
  
  if      ((diff_x ==  0) && (diff_y ==  1))
    {
      printf("follow_path(): Moving North.\n");
      sprintf(cmd, "1 Move N\n");
    }
  else if ((diff_x ==  0) && (diff_y == -1))
    {
      printf("follow_path(): Moving South.\n");
      sprintf(cmd, "1 Move S\n");
    }
  else if ((diff_x ==  1) && (diff_y ==  0))
    {
      printf("follow_path(): Moving East.\n");
      sprintf(cmd, "1 Move E\n");
    }
  else if ((diff_x == -1) && (diff_y == 0))
    {
      printf("follow_path(): Moving West.\n");
      sprintf(cmd, "1 Move W\n");
    }
  else
    {
      printf("!!**!! Somehow I'm off the path ... Replanning.\n");

      /* Execute a no-op this turn.
       */
      if (*path == NULL)
        {
          /* At the end of the path already
           */
          sprintf(cmd, "1 Pick\n");
          return(OK);
        }

      /* Find the destination node
       */
      tmp = *path;
      while (tmp->next != NULL)
        {
          tmp = tmp->next;
        }
      dest_x = tmp->node->x;
      dest_y = tmp->node->y;

      /* Blow away the current path
       */
      path_delete(path);

      /* Plan a new path
       */
      printf("Planning Path from (%d, %d) to (%d, %d)\n",
             robot_x, robot_y, dest_x, dest_y);

      status = plan_path(robot_x, robot_y, dest_x, dest_y, path);
      if (status != OK)
        {
          fprintf(stderr, "follow_path(): Error line: %d\n", __LINE__);
          return(ERROR);
        }

#ifdef DEBUG_PF
      print_path(*path);
#endif
      
      /* Update the game state
       */
      game_state.active_path = *path;
      
      /* First node is always our start position (Remove it).
       */
      status = path_pop_node(&(game_state.active_path), &node);
      if (status != OK)
        {
          fprintf(stderr, "follow_path(): Error in path_pop_node().\n");
          return(ERROR);
        }

      if (path_length(game_state.active_path) >= 1)
        {
          status = path_pop_node(path, &node);
          if (status != OK)
            {
              fprintf(stderr, "follow_path(): Error in path_pop_node().\n");
              return(ERROR);
            }

          new_x = node->x;
          new_y = node->y;
          
          diff_x = new_x - robot_x;
          diff_y = new_y - robot_y;
          
          if      ((diff_x ==  0) && (diff_y ==  1))
            {
              printf("follow_path(): Moving North.\n");
              sprintf(cmd, "1 Move N\n");
            }
          else if ((diff_x ==  0) && (diff_y == -1))
            {
              printf("follow_path(): Moving South.\n");
              sprintf(cmd, "1 Move S\n");
            }
          else if ((diff_x ==  1) && (diff_y ==  0))
            {
              printf("follow_path(): Moving East.\n");
              sprintf(cmd, "1 Move E\n");
            }
          else if ((diff_x == -1) && (diff_y == 0))
            {
              printf("follow_path(): Moving West.\n");
              sprintf(cmd, "1 Move W\n");
            }
        }
      else
        {
          /* We're sitting on our destination already. No-Op.
           */
          sprintf(cmd, "1 Drop\n");
        }

    } /* End Replan */

  return(OK);
}


/***************************************************
 * find_closest_home()                             *
 *                                                 *
 ***************************************************/
static int find_closest_home(int *x, int *y)
{
  int robot_x;
  int robot_y;
  int width;
  int height;
  int dist;
  int lowest_dist;
  int found;
  int i;
  int j;

  width   = game_state.map_size_x;
  height  = game_state.map_size_y;
  robot_x = game_state.robot_x;
  robot_y = game_state.robot_y;

  /* Init lowest_dist to the max grid travel distance
   */
  lowest_dist = (width + height);

  /* Find the closest home base to our current position
   *
   * This isn't the best, since we're not planning a 
   * path to each of the possibilities, and taking that
   * into account, but it will at least give us an
   * initial goal.
   */
  found = FALSE;
  for (i = 0; i < height; i++)
    {
      for (j = 0; j < width; j++)
        {
          if (game_map[i][j].type == TERRAIN_HOME)
            {
              /* Found a home base square!
               * (That we haven't seen before)
               */
              found = TRUE;

              dist = goal_dist_est(robot_x,robot_y,i,j);

              if (game_map[i][j].visited)
                {
                  /* Slightly dislike those we've visited already
                   */
                  dist += (game_map[i][j].visited * 2);

                  /* Seriously dislike for those we've seen that
                   * have no packages ...
                   */
                  if (game_map[i][j].packages == 0)
                    {
                      dist += (game_map[i][j].visited * 10);
                    }
                }

              if (dist < lowest_dist)
                {
                  lowest_dist = dist;
                  *x = j;
                  *y = i;
                }
            }
        }
    }

  if (found)
    {
      /* Increment the visit count for this node.
       */
      game_map[*y][*x].visited++;
      return(OK);
    }
  return(ERROR);
}


/***************************************************
 * plan_path()                                     *
 *                                                 *
 ***************************************************/
static int plan_path(int start_x, int start_y, int dest_x, int dest_y,
                     Path_entry **path)
{
  /* Plan a path from Starting Location (start_x, start_y)
   * to Destination Location (dest_x, dest_y)
   *
   * This code is based on the pseudo-code presented in
   * Gamasutra, which can be found at:
   *
   * http://www.gamasutra.com/features/19970801/pathfinding.html
   *
   * However all this C code is written from scratch,
   * including the ugly linked list manipulation code ;)
   *
   */
  Map_node       *s;      /* Pointer to the start map node         */
  Map_node       *n;      /* Pointer to a generic map node         */
  Map_node       *np;     /* Pointer to a connected map node       */
  Pf_pque_entry  *open;   /* Priority que of open map nodes (head) */
  Pf_list_entry  *closed; /* List of closed map nodes (head)       */
  int             il_break;
  int             status;
  char            errmsg[] = "plan_path(): Error line";
  int             newg;
  int             in_open;
  int             in_closed;
  int             better_score;
  int             i;

  /* Init Variables
   */
  s         = &(game_map[start_y][start_x]);
  open      = NULL;
  closed    = NULL;
  il_break  = FALSE;

  s->g      = 0;
  s->h      = goal_dist_est(start_x, start_y, dest_x, dest_y);
  s->f      = (s->g) + (s->h);
  s->parent = NULL;


  /* Push the starting node
   */
  status = pque_push_node(&open, s);
  if (status != OK)
    {
      fprintf(stderr, "%s: %d.\n", errmsg, __LINE__);
      pque_delete(&open);
      return(ERROR);
    }

  while (!pque_is_empty(open))
    {
      status = pque_pop_node(&open, &n);
      if (status != OK)
        {
          fprintf(stderr, "%s: %d.\n", errmsg, __LINE__);
          break;
        }

      /* Check if node N is a goal node
       */
      if ((n->x == dest_x) && (n->y == dest_y))
        {
#ifdef DEBUG_PF
          printf("plan_path(): Goal node found! Success!\n");
#endif
          /* Construct path
           */
          while (n != NULL)
            {
#ifdef DEBUG_PF
              printf("n->x: %d n->y: %d\n", n->x, n->y);
#endif
              status = path_push_node(path, n);
              if (status != OK)
                {
                  fprintf(stderr, "%s: %d.\n", errmsg, __LINE__);
                  break;
                }
              n = n->parent;
            }

          /* Cleanup
           */
          pque_delete(&open);
          list_delete(&closed);
          
          /* Return success
           */
          return(OK);
        }
      else
        {
#ifdef DEBUG_PF
          printf("plan_path(): Node x: %d y: %d is not the goal.\n",
                 n->x, n->y);
#endif
        }

      /* Not the goal yet. Check connected nodes.
       */
      for (i = 0; i < 4; i++)
        {
          if (n->links & (1 << i))
            {
              /* Connected node found
               */
              switch (i)
                {
                case 0: /* North */
#ifdef DEBUG_PF
                  printf("North is connected.\n");
#endif
                  np = &(game_map[n->y+1][n->x]);
                  break;
                case 1: /* South */
#ifdef DEBUG_PF
                  printf("South is connected.\n");
#endif
                  np = &(game_map[n->y-1][n->x]);
                  break;
                case 2: /* East  */
#ifdef DEBUG_PF
                  printf("East is connected.\n");
#endif
                  np = &(game_map[n->y][n->x+1]);
                  break;
                case 3: /* West  */
#ifdef DEBUG_PF
                  printf("West is connected.\n");
#endif
                  np = &(game_map[n->y][n->x-1]);
                  break;
                }

              /* cost(n,n') is always 1, for now.
               */
              newg = (n->g) + 1;

#ifdef DEBUG_PF
              printf("Finding node in Open ...\n");
#endif
              in_open      = pque_find_node(open,   np);

#ifdef DEBUG_PF
              printf("Finding node in Closed ...\n");
#endif
              in_closed    = list_find_node(closed, np);

              better_score = (np->g) <= newg;

#ifdef DEBUG_PF
              printf("newg: %d  in_open: %d in_closed: %d better_score: %d\n",
                     newg, in_open, in_closed, better_score);
#endif

              if ( (in_open || in_closed) && (better_score) )
                {
                  /* Do nothing */
                }
              else
                {
                  np->parent = n;
                  np->g      = newg;
                  np->h      = goal_dist_est(np->x, np->y, dest_x, dest_y);
                  np->f      = (np->g) + (np->h);
                  
                  if (in_closed)
                    {
                      status = list_remove_node(&closed, np);
                      if (status != OK)
                        {
                          fprintf(stderr, "%s: %d.\n", errmsg, __LINE__);
                          il_break = TRUE;
                          break;
                        }
                    }

                  if (!in_open)
                    {
                      status = pque_push_node(&open, np);
                      if (status != OK)
                        {
                          fprintf(stderr, "%s: %d.\n", errmsg, __LINE__);
                          il_break = TRUE;
                          break;
                        }
                    }

                } /* If we're not skipping the node */

            } /* If Direction Connected */

        } /* Loop over links */

      if (il_break)
        {
          /* Break requested from the inner loop */
          break;
        }

      status = list_push_node(&closed, n);
      if (status != OK)
        {
          fprintf(stderr, "%s: %d.\n", errmsg, __LINE__);
          break;
        }
    }

  /* Cleanup
   */
  pque_delete(&open);
  list_delete(&closed);

  return(ERROR); /* No path found */
}


/***************************************************
 * goal_dist_est()                                 *
 *                                                 *
 ***************************************************/
int goal_dist_est(int s_x, int s_y, int d_x, int d_y)
{
  /* Take the difference in x & y position between the
   * start and the destination positions
   */
  return(abs(d_x-s_x) + abs(d_y-s_y));
}


/***************************************************
 * path_length()                                   *
 *                                                 *
 ***************************************************/
static int path_length(Path_entry *head)
{
  Path_entry *tmp;
  int         count;

  count = 0;

  tmp = head;
  while (tmp != NULL)
    {
      count++;
      tmp = tmp->next;
    }

  return(count);
}


/***************************************************
 * path_push_node()                                *
 *                                                 *
 ***************************************************/
static int path_push_node(Path_entry **head, Map_node *node)
{
  Path_entry *new_entry;

  /* Allocate an entry
   */
  new_entry = (Path_entry *) malloc(sizeof(Path_entry));
  if (new_entry == NULL)
    {
      fprintf(stderr, "path_push_node(): Error in malloc().\n");
      return(ERROR);
    }

  new_entry->node = node;
  new_entry->next = NULL;
  new_entry->prev = NULL;

  if (*head == NULL)
    {
      *head = new_entry;
    }
  else
    {
      /* Since paths are traversed from the planner in reverse order,
       * we always push them onto the head of the list, so that it
       * comes out forward.
       */
      new_entry->next = *head;
      (*head)->prev   =  new_entry;
      *head           =  new_entry;
    }

#ifdef DEBUG_PF
  printf("path_push_node(): Pushed node: x: %d y: %d\n",
         node->x, node->y);
#endif

  return(OK);
}


/***************************************************
 * path_pop_node()                                 *
 *                                                 *
 ***************************************************/
static int path_pop_node(Path_entry **head, Map_node **node)
{
  Path_entry *tmp;

  if (*head == NULL)
    {
      fprintf(stderr, "path_pop_node(): Nothing to pop!\n");
      return(ERROR);
    }

  /* Get the node
   */
  *node = (*head)->node;

  tmp   = *head;

  *head = tmp->next;
  if (*head != NULL)
    {
      (*head)->prev = NULL;
    }

  free(tmp);

#ifdef DEBUG_PF
  printf("path_pop_node(): Popped node x: %d y: %d\n", (*node)->x, (*node)->y);
#endif

  return(OK);
}


/***************************************************
 * pque_push_node()                                *
 *                                                 *
 ***************************************************/
int pque_push_node(Pf_pque_entry **head, Map_node *node)
{
  Pf_pque_entry *new_entry;
  Pf_pque_entry *tmp;

  /* Allocate an entry
   */
  new_entry = (Pf_pque_entry *) malloc(sizeof(Pf_pque_entry));
  if (new_entry == NULL)
    {
      fprintf(stderr, "pque_push_node(): Error in malloc().\n");
      return(ERROR);
    }

  new_entry->node = node; /* Save the address of the node */
  new_entry->prev = NULL;
  new_entry->next = NULL;

  if (*head == NULL)
    {
      *head = new_entry;
    }
  else
    {
      /* Push the new entry onto the end of the list.  Instead of
       * trying to sort on insert, we'll do the search on the pop.
       */
      tmp = *head;
      while (tmp->next != NULL)
        {
          tmp = tmp->next;
        }

      tmp->next       = new_entry;
      new_entry->prev = tmp;
    }

#ifdef DEBUG_PF
  printf("pque_push_node(): Pushed node x: %d y: %d\n", node->x, node->y);
#endif

  return(OK);
}


/***************************************************
 * list_push_node()                                *
 *                                                 *
 ***************************************************/
static int list_push_node(Pf_list_entry **head, Map_node *node)
{
  return( pque_push_node((Pf_pque_entry **) head, node) );
}


/***************************************************
 * pque_pop_node()                                 *
 *                                                 *
 ***************************************************/
static int pque_pop_node(Pf_pque_entry **head, Map_node **node)
{
  Pf_list_entry *prev;
  Pf_list_entry *next;
  Pf_pque_entry *tmp;
  Pf_pque_entry *tmp_save;
  int            best_score;

  if (*head == NULL)
    {
      fprintf(stderr, "pque_pop_node(): Nothing to pop!\n");
      return(ERROR);
    }

  tmp        = *head;
  best_score = 2147483647; /* 2^31 - 1 */
  while (tmp != NULL)
    {
      if (tmp->node->f < best_score)
        {
          best_score = tmp->node->f;
          *node      = tmp->node;
          tmp_save   = tmp;
        }

      tmp = tmp->next;
    }

  /* Remove the best cost pf_pque_entry
   */
  if ((tmp_save->prev == NULL) || (tmp_save->next == NULL))
    {
      /* Special Cases
       */
      if (tmp_save->prev == NULL)
        {
          /* Removed node is the head of the pque
           */
          *head = tmp_save->next;
          if (*head != NULL)
            {
              /* i.e. More than 1 entry in pque
               */
              (*head)->prev = NULL;
            }
        }
      else
        {
          /* Removed node is the tail of the pque
           */
          (tmp_save->prev)->next = NULL;
        }
    }
  else
    {
      /* General Case
       */
      prev = tmp_save->prev;
      next = tmp_save->next;
      
      prev->next = next;
      next->prev = prev;
    }

  free(tmp_save);

#ifdef DEBUG_PF
  printf("pque_pop_node(): Popped node x: %d y: %d\n", (*node)->x, (*node)->y);
#endif

  return(OK);
}


/***************************************************
 * list_remove_node()                              *
 *                                                 *
 ***************************************************/
static int list_remove_node(Pf_list_entry **head, Map_node *node)
{
  Pf_list_entry *prev;
  Pf_list_entry *next;
  Pf_list_entry *tmp;
  int            node_x;
  int            node_y;
  int            found;

  if (*head == NULL)
    {
      fprintf(stderr, "list_remove_node(): Nothing to remove!\n");
      return(ERROR);
    }

  /* Find the desired node
   */
  tmp    = *head;
  node_x = node->x;
  node_y = node->y;
  found  = FALSE;
  while (tmp != NULL)
    {
      if ((tmp->node->x == node_x) && (tmp->node->y == node_y))
        {
          /* Found the node
           */
          found = TRUE;
          break;
        }
    }
  
  if (!found)
    {
      fprintf(stderr, "list_remove_node(): Node not found!\n");
      return(ERROR);
    }

  /* Delete the entry from the list
   */
  if ((tmp->prev == NULL) || (tmp->next == NULL))
    {
      /* Special Cases
       */
      if (tmp->prev == NULL)
        {
          /* Removed node is the head of the list
           */
          *head = tmp->next;
          if (*head != NULL)
            {
              /* i.e. More than 1 entry in list
               */
              (*head)->prev = NULL;
            }
        }
      else
        {
          /* Removed node is the tail of the list
           */
          (tmp->prev)->next = NULL;
        }
    }
  else
    {
      /* General Case
       */
      prev = tmp->prev;
      next = tmp->next;
      
      prev->next = next;
      next->prev = prev;
    }

  free(tmp);

#ifdef DEBUG_PF
  printf("list_remove_node(): Removed node x: %d y: %d from list\n",
         node->x, node->y);
#endif
  
  return(OK);
}


/***************************************************
 * pque_find_node()                                *
 *                                                 *
 ***************************************************/
static int pque_find_node(Pf_pque_entry *head, Map_node *node)
{
  int            node_x;
  int            node_y;
  Pf_pque_entry *tmp;
  
  if (head == NULL)
    {
      return(FALSE);
    }

  tmp    = head;
  node_x = node->x;
  node_y = node->y;
  while (tmp != NULL)
    {
      if ((tmp->node->x == node_x) &&
          (tmp->node->y == node_y))
        {
          return(TRUE);
        }

      tmp = tmp->next;
    }

  return(FALSE);
}


/***************************************************
 * list_find_node()                                *
 *                                                 *
 ***************************************************/
static int list_find_node(Pf_list_entry *head, Map_node *node)
{
  return ( pque_find_node((Pf_list_entry *) head, node) );
}


/***************************************************
 * pque_is_empty()                                 *
 *                                                 *
 ***************************************************/
static int pque_is_empty(Pf_pque_entry *head)
{
  if (head == NULL)
    {
      return(TRUE);
    }
  return(FALSE);
}


/**************************************************
 * pque_delete()                                  *
 *                                                *
 **************************************************/
static void pque_delete(Pf_pque_entry **head)
{
  Pf_pque_entry *tmp;

  while (*head != NULL)
    {
      tmp = (*head)->next;

      free(*head);

      *head = tmp;
    }
}


/**************************************************
 * list_delete()                                  *
 *                                                *
 **************************************************/
static void list_delete(Pf_list_entry **head)
{
  return( pque_delete((Pf_pque_entry **) head) );
}


/**************************************************
 * path_delete()                                  *
 *                                                *
 **************************************************/
static void path_delete(Path_entry **head)
{
  return( pque_delete((Pf_pque_entry **) head) );
}

/************************************************************
 * Socket Functions                                         *
 ************************************************************/

/***************************************************
 * client_connect()                                *
 *                                                 *
 ***************************************************/
static int client_connect(char *host_str, char *port_str)
{
  struct sockaddr_in server;
  int                sp0;
  socklen_t          len;
  unsigned short     port;
  int                status;
  struct hostent    *hp;

  /* Handle creating a TCP/IP socket connection to the server
   */

  /* Create a socket
   */
  game_state.socket_open = FALSE;
  sp0 = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (sp0 < 0)
    {
      perror("socket(): Error ");
      fprintf(stderr, "client_connect(): Error in socket().\n");
      return(ERROR);
    }
  game_state.socket_open = TRUE;

  /* Save the socket handle
   */
  game_state.socket = sp0;


  /* Clear the server structure
   */
  memset((char *) &(server), 0x0, sizeof(server));

  /* Set fields in the server structure
   */
  server.sin_family = AF_INET;

  /* Convert port number to network byte order
   */
  port = (unsigned short) atoi(port_str);
  server.sin_port = htons(port);

  /* Convert numbers-and-dots notation into
   * binary data in network byte order
   */
  if (isdigit(host_str[0]))
    {
      /* Handle numeric address
       */
      server.sin_addr.s_addr = inet_addr(host_str);
    }
  else
    {
      printf("Attempting to resolve host name: %s\n", host_str);
      
      if ((hp = gethostbyname(host_str)) == NULL)
        {
          perror("gethostbyname(): ");
          return(ERROR);
        }

      server.sin_addr.s_addr = ((struct in_addr *) (hp->h_addr))->s_addr;
    }

  printf("Attempting to connect to: %s port: %s\n",
         inet_ntoa(server.sin_addr), port_str);

  /* Connect to the server
   */
  len = (socklen_t) sizeof(struct sockaddr_in);
  status = connect(sp0, (struct sockaddr *) &server, len);
  if (status < 0)
    {
      perror("connect(): Error ");
      fprintf(stderr, "client_connect(): Error in connect().\n");
      return(ERROR);
    }

  return(OK);
}


/***************************************************
 * socket_fgets()                                  *
 *                                                 *
 ***************************************************/
static int socket_fgets(char *data, int max_bytes)
{
  char tmp;
  int  i;
  int  sp0;
  int  bytes_read;

  sp0 = game_state.socket;

  for (i = 0; i < (max_bytes-1); i++)
    {
      bytes_read = read(sp0, &tmp, 1);

      if (bytes_read != 1)
        {
          perror("socket_fgets: Error ");
          fprintf(stderr, "socket_fgets(): Error in read().\n");
          return(ERROR);
        }

      data[i] = tmp;

      if (tmp == '\n')
        {
          i++;
          break;
        }
    }

  data[i] = '\0';

  return(OK);
}

/***************************************************
 * socket_fgets_unlim()                            *
 *                                                 *
 ***************************************************/
static int socket_fgets_unlim(char **pbuffer)
{
#define BUFF_INC 1024

  int   buffer_size;
  char *buffer;
  int   eol_found;
  char *ptmp;
  int   status;

  
  /* Allocate memory and read from socket until we get
   * an EOL character, reallocating memory as needed.
   */
  *pbuffer = NULL;


  /* Allocate an initial amount of memory for the buffer, assume BUFF_INC
   */
  buffer_size = BUFF_INC;
  buffer = (char *) malloc(buffer_size * sizeof(char));
  if (buffer == NULL)
    {
      fprintf(stderr, "socket_fgets_unlim(): Error in malloc().\n");
      return(ERROR);
    }


  /* Read into the buffer until we find a newline character.
   * If needed, allocate more memory.
   */
  eol_found = FALSE;
  ptmp      = &(buffer[0]);

  while (!eol_found)
    {
      status = socket_fgets(ptmp, BUFF_INC);
      if (status != OK)
        {
          fprintf(stderr, "socket_fgets_unlim(): Error in socket_fgets().\n");
          return(ERROR);
        }

      if (strchr(ptmp, '\n') != NULL)
        {
          eol_found = TRUE;
        }
      else
        {
          printf("Allocating more buffer space.\n");

          /* Allocate more space ... long line.
           */
          buffer_size += BUFF_INC;

          buffer = (char *) realloc(buffer, buffer_size * sizeof(char *));
          if (errno == ENOMEM)
            {
              fprintf(stderr, "socket_fgets_unlim(): Error in ralloc()\n");
              free(buffer);
              return(ERROR);
            }

          if (buffer == NULL)
            {
              fprintf(stderr, "socket_fgets_unlim(): Error in realloc()\n");
              free(buffer);
              return(ERROR);
            }

          ptmp += (BUFF_INC-1);
        }
    }

  /* Set return pointer
   */
  *pbuffer = buffer;

  return(OK);
}



/***************************************************
 * socket_write()                                  *
 *                                                 *
 ***************************************************/
static int socket_write(char *data, int num_bytes)
{
  int bytes_written;
  int sp0;

  /* Get socket
   */
  sp0 = game_state.socket;

  bytes_written = write(sp0, data, num_bytes);

  if (bytes_written != num_bytes)
    {
      perror("write(): Error ");
      fprintf(stderr, "socket_write(): Error in write().\n");
      return(ERROR);
    }

  return(OK);
}


/***************************************************
 * socket_read()                                   *
 *                                                 *
 ***************************************************/
static int socket_read(char *data, int num_bytes)
{
  int   bytes_left;
  int   bytes_read;
  int   sp0;
  char *tmp;

  sp0 = game_state.socket;

  bytes_left = num_bytes;
  tmp        = data;

  while (bytes_left > 0)
    {
      bytes_read = read(sp0, tmp, bytes_left);

      if (bytes_read < 0)
        {
          perror("read(): Error ");
          fprintf(stderr, "socket_read(): Error from read().\n");
          return(ERROR);
        }

      bytes_left -= bytes_read;
      tmp        += bytes_read;
    }

  return(OK);
}
